"""A list of all supported languages in highlight.js.

This file is automatically generated by tools.
"""
import typing


class Language(typing.NamedTuple):
    """A language supported by highlight.js."""
    name: str
    alias: str


languages: typing.List[Language] = [
    Language(name='1C:Enterprise', alias='1c'),
    Language(name='Augmented Backus-Naur Form', alias='abnf'),
    Language(name='Apache Access Log', alias='accesslog'),
    Language(name='ActionScript', alias='actionscript'),
    Language(name='Ada', alias='ada'),
    Language(name='AngelScript', alias='angelscript'),
    Language(name='Apache config', alias='apache'),
    Language(name='AppleScript', alias='applescript'),
    Language(name='ArcGIS Arcade', alias='arcade'),
    Language(name='Arduino', alias='arduino'),
    Language(name='ARM Assembly', alias='armasm'),
    Language(name='AsciiDoc', alias='asciidoc'),
    Language(name='AspectJ', alias='aspectj'),
    Language(name='AutoHotkey', alias='autohotkey'),
    Language(name='AutoIt', alias='autoit'),
    Language(name='AVR Assembly', alias='avrasm'),
    Language(name='Awk', alias='awk'),
    Language(name='X++', alias='axapta'),
    Language(name='Bash', alias='bash'),
    Language(name='BASIC', alias='basic'),
    Language(name='Backus–Naur Form', alias='bnf'),
    Language(name='Brainfuck', alias='brainfuck'),
    Language(name='C', alias='c'),
    Language(name='C/AL', alias='cal'),
    Language(name='Cap’n Proto', alias='capnproto'),
    Language(name='Ceylon', alias='ceylon'),
    Language(name='Clean', alias='clean'),
    Language(name='Clojure', alias='clojure'),
    Language(name='Clojure REPL', alias='clojure-repl'),
    Language(name='CMake', alias='cmake'),
    Language(name='CoffeeScript', alias='coffeescript'),
    Language(name='Coq', alias='coq'),
    Language(name='Caché Object Script', alias='cos'),
    Language(name='C++', alias='cpp'),
    Language(name='crmsh', alias='crmsh'),
    Language(name='Crystal', alias='crystal'),
    Language(name='C#', alias='csharp'),
    Language(name='CSP', alias='csp'),
    Language(name='CSS', alias='css'),
    Language(name='D', alias='d'),
    Language(name='Dart', alias='dart'),
    Language(name='Delphi', alias='delphi'),
    Language(name='Diff', alias='diff'),
    Language(name='Django', alias='django'),
    Language(name='DNS Zone', alias='dns'),
    Language(name='Dockerfile', alias='dockerfile'),
    Language(name='Batch file (DOS)', alias='dos'),
    Language(name='dsconfig', alias='dsconfig'),
    Language(name='Device Tree', alias='dts'),
    Language(name='Dust', alias='dust'),
    Language(name='Extended Backus-Naur Form', alias='ebnf'),
    Language(name='Elixir', alias='elixir'),
    Language(name='Elm', alias='elm'),
    Language(name='ERB', alias='erb'),
    Language(name='Erlang', alias='erlang'),
    Language(name='Erlang REPL', alias='erlang-repl'),
    Language(name='Excel formulae', alias='excel'),
    Language(name='FIX', alias='fix'),
    Language(name='Flix', alias='flix'),
    Language(name='Fortran', alias='fortran'),
    Language(name='F#', alias='fsharp'),
    Language(name='GAMS', alias='gams'),
    Language(name='GAUSS', alias='gauss'),
    Language(name='G-code (ISO 6983)', alias='gcode'),
    Language(name='Gherkin', alias='gherkin'),
    Language(name='GLSL', alias='glsl'),
    Language(name='GML', alias='gml'),
    Language(name='Go', alias='go'),
    Language(name='Golo', alias='golo'),
    Language(name='Gradle', alias='gradle'),
    Language(name='GraphQL', alias='graphql'),
    Language(name='Groovy', alias='groovy'),
    Language(name='HAML', alias='haml'),
    Language(name='Handlebars', alias='handlebars'),
    Language(name='Haskell', alias='haskell'),
    Language(name='Haxe', alias='haxe'),
    Language(name='HSP', alias='hsp'),
    Language(name='HTTP', alias='http'),
    Language(name='Hy', alias='hy'),
    Language(name='Inform 7', alias='inform7'),
    Language(name='TOML, also INI', alias='ini'),
    Language(name='IRPF90', alias='irpf90'),
    Language(name='ISBL', alias='isbl'),
    Language(name='Java', alias='java'),
    Language(name='JavaScript', alias='javascript'),
    Language(name='JBoss CLI', alias='jboss-cli'),
    Language(name='JSON', alias='json'),
    Language(name='Julia', alias='julia'),
    Language(name='Julia REPL', alias='julia-repl'),
    Language(name='Kotlin', alias='kotlin'),
    Language(name='Lasso', alias='lasso'),
    Language(name='LaTeX', alias='latex'),
    Language(name='LDIF', alias='ldif'),
    Language(name='Leaf', alias='leaf'),
    Language(name='Less', alias='less'),
    Language(name='Lisp', alias='lisp'),
    Language(name='LiveCode', alias='livecodeserver'),
    Language(name='LiveScript', alias='livescript'),
    Language(name='LLVM IR', alias='llvm'),
    Language(name='LSL (Linden Scripting Language)', alias='lsl'),
    Language(name='Lua', alias='lua'),
    Language(name='Makefile', alias='makefile'),
    Language(name='Markdown', alias='markdown'),
    Language(name='Mathematica', alias='mathematica'),
    Language(name='Matlab', alias='matlab'),
    Language(name='Maxima', alias='maxima'),
    Language(name='MEL', alias='mel'),
    Language(name='Mercury', alias='mercury'),
    Language(name='MIPS Assembly', alias='mipsasm'),
    Language(name='Mizar', alias='mizar'),
    Language(name='Mojolicious', alias='mojolicious'),
    Language(name='Monkey', alias='monkey'),
    Language(name='MoonScript', alias='moonscript'),
    Language(name='N1QL', alias='n1ql'),
    Language(name='Nested Text', alias='nestedtext'),
    Language(name='Nginx config', alias='nginx'),
    Language(name='Nim', alias='nim'),
    Language(name='Nix', alias='nix'),
    Language(name='Node REPL', alias='node-repl'),
    Language(name='NSIS', alias='nsis'),
    Language(name='Objective-C', alias='objectivec'),
    Language(name='OCaml', alias='ocaml'),
    Language(name='OpenSCAD', alias='openscad'),
    Language(name='Oxygene', alias='oxygene'),
    Language(name='Parser3', alias='parser3'),
    Language(name='Perl', alias='perl'),
    Language(name='Packet Filter config', alias='pf'),
    Language(name='PostgreSQL', alias='pgsql'),
    Language(name='php', alias='php'),
    Language(name='PHP template', alias='php-template'),
    Language(name='Plain text', alias='plaintext'),
    Language(name='Pony', alias='pony'),
    Language(name='PowerShell', alias='powershell'),
    Language(name='Processing', alias='processing'),
    Language(name='Python profiler', alias='profile'),
    Language(name='Prolog', alias='prolog'),
    Language(name='.properties', alias='properties'),
    Language(name='Protocol Buffers', alias='protobuf'),
    Language(name='Puppet', alias='puppet'),
    Language(name='PureBASIC', alias='purebasic'),
    Language(name='Python', alias='python'),
    Language(name='python-repl', alias='python-repl'),
    Language(name='Q', alias='q'),
    Language(name='QML', alias='qml'),
    Language(name='R', alias='r'),
    Language(name='ReasonML', alias='reasonml'),
    Language(name='RenderMan RIB', alias='rib'),
    Language(name='Roboconf', alias='roboconf'),
    Language(name='MikroTik RouterOS script', alias='routeros'),
    Language(name='RenderMan RSL', alias='rsl'),
    Language(name='Ruby', alias='ruby'),
    Language(name='Oracle Rules Language', alias='ruleslanguage'),
    Language(name='Rust', alias='rust'),
    Language(name='SAS', alias='sas'),
    Language(name='Scala', alias='scala'),
    Language(name='Scheme', alias='scheme'),
    Language(name='Scilab', alias='scilab'),
    Language(name='SCSS', alias='scss'),
    Language(name='Shell Session', alias='shell'),
    Language(name='Smali', alias='smali'),
    Language(name='Smalltalk', alias='smalltalk'),
    Language(name='SML (Standard ML)', alias='sml'),
    Language(name='SQF', alias='sqf'),
    Language(name='SQL', alias='sql'),
    Language(name='Stan', alias='stan'),
    Language(name='Stata', alias='stata'),
    Language(name='STEP Part 21', alias='step21'),
    Language(name='Stylus', alias='stylus'),
    Language(name='SubUnit', alias='subunit'),
    Language(name='Swift', alias='swift'),
    Language(name='Tagger Script', alias='taggerscript'),
    Language(name='Test Anything Protocol', alias='tap'),
    Language(name='Tcl', alias='tcl'),
    Language(name='Thrift', alias='thrift'),
    Language(name='TP', alias='tp'),
    Language(name='Twig', alias='twig'),
    Language(name='TypeScript', alias='typescript'),
    Language(name='Vala', alias='vala'),
    Language(name='Visual Basic .NET', alias='vbnet'),
    Language(name='VBScript', alias='vbscript'),
    Language(name='VBScript in HTML', alias='vbscript-html'),
    Language(name='Verilog', alias='verilog'),
    Language(name='VHDL', alias='vhdl'),
    Language(name='Vim Script', alias='vim'),
    Language(name='WebAssembly', alias='wasm'),
    Language(name='Wren', alias='wren'),
    Language(name='Intel x86 Assembly', alias='x86asm'),
    Language(name='XL', alias='xl'),
    Language(name='HTML, XML', alias='xml'),
    Language(name='XQuery', alias='xquery'),
    Language(name='YAML', alias='yaml'),
    Language(name='Zephir', alias='zephir'),
]